/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Application
    blockMesh

Description
    Mesh generator

\*---------------------------------------------------------------------------*/

#include "blockMesh.H"

#include "HashSet.H"


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

// Construct from IOdictionary
blockMesh::blockMesh(IOdictionary& meshDescription)
:
    topologyPtr_(createTopology(meshDescription)),
    scale_(readScalar(meshDescription.lookup("convertToMeters"))),
    blockOffsets_(createBlockOffsets()),
    mergeList_(createMergeList()),
    points_(createPoints()),
    cells_(createCells()),
    patches_(createPatches())
{}


// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

blockMesh::~blockMesh()
{
    delete topologyPtr_;
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

const polyMesh& blockMesh::topology() const
{
    if (!topologyPtr_)
    {
        FatalErrorIn("blockMesh::topology() const")
            << "topologyPtr_ not allocated"
            << exit(FatalError);
    }

    return *topologyPtr_;
}


const curvedEdgeList& blockMesh::edges() const
{
    return edges_;
}

wordList blockMesh::patchNames() const
{
    const polyPatchList& patchTopologies = topology().boundaryMesh();
    wordList names(patchTopologies.size());

    forAll (names, patchI)
    {
        names[patchI] = patchTopologies[patchI].name();
    }

    return names;
}


wordList blockMesh::patchTypes() const
{
    const polyPatchList& patchTopologies = topology().boundaryMesh();
    wordList types(patchTopologies.size());

    forAll (types, patchI)
    {
        types[patchI] = patchTopologies[patchI].type();
    }

    return types;
}


wordList blockMesh::patchPhysicalTypes() const
{
    const polyPatchList& patchTopologies = topology().boundaryMesh();
    wordList physicalTypes(patchTopologies.size());

    forAll (physicalTypes, patchI)
    {
        physicalTypes[patchI] = patchTopologies[patchI].physicalType();
    }

    return physicalTypes;
}

wordList blockMesh::cellSetNames() const
{
  HashSet<word> names;
  forAll((*this),b) {
    word name=(*this)[b].blockDef().cellSetName();
    if(name!="") {
      if(!names.found(name)) {
	names.insert(name);
      }
    }
  }

  return names.toc();
}

void blockMesh::addCellZones(polyMesh &mesh) const
{
  labelList offsets((*this).size());
  wordList names=cellSetNames();

  offsets[0]=0;

  forAll((*this),blockI) {
    block b=(*this)[blockI];
    if(blockI>0) {
      offsets[blockI]=offsets[blockI-1]+b.cells().size();
    }
  }
  
  List<pointZone*> pz(0);

  //  List<faceZone*> fz(0);

  // create a dummy-face Zone so that the cellZones are written
  List<faceZone*> fz(1);
  fz[0]=new faceZone(
		     word("dummy"),
		     labelList(0),
		     boolList(0,false),
		     666,
		     mesh.faceZones()
		     );

  List<cellZone*> cz(names.size());

  forAll(names,nI) {
    word name=names[nI];

    label nr=0;
    forAll((*this),blockI) {
      block b=(*this)[blockI];
      if(b.blockDef().cellSetName()==name) {
	nr+=b.cells().size();
      }
    }

    Info << "Adding region " << name << " with " << nr << " cells" << endl;

    labelList cls(nr);
    label cnt=0;

    forAll((*this),blockI) {
      block b=(*this)[blockI];
      if(b.blockDef().cellSetName()==name) {
	labelListList cells=b.cells();
	label off=offsets[blockI];
	label anz=b.cells().size();

	for(label i=off;i<off+anz;i++) {
	  cls[cnt]=i;
	  cnt++;
	}
      }
    }

    cz[nI]=new cellZone
      (
       name,
       cls,
       nI,
       mesh.cellZones()
       );

  }

  mesh.addZones(pz,fz,cz);
}

void blockMesh::writeCellSet(const polyMesh &mesh,const word &name) const
{
  label nr=0;
  labelList offsets((*this).size());
  offsets[0]=0;

  forAll((*this),blockI) {
    block b=(*this)[blockI];
    if(b.blockDef().cellSetName()==name) {
      nr+=b.cells().size();
    }
    if(blockI>0) {
      offsets[blockI]=offsets[blockI-1]+b.cells().size();
    }
  }

  Info << "CellSet " << name << " with " << nr << " cells" << endl;

  cellSet cs(mesh,name,nr);

  forAll((*this),blockI) {
    block b=(*this)[blockI];
    if(b.blockDef().cellSetName()==name) {
      labelListList cells=b.cells();
      label off=offsets[blockI];
      label anz=b.cells().size();

      //      Info << name << " : " << off << " " << anz << endl;

      for(label i=off;i<off+anz;i++) {
	cs.insert(i);
      }
    }
  }
  cs.write();
}

void blockMesh::writeTopology(Ostream& os) const
{
    const pointField& pts = topology().points();

    forAll(pts, pI)
    {
        const point& pt = pts[pI];

        os << "v " << pt.x() << ' ' << pt.y() << ' ' << pt.z() << endl;
    }

    const edgeList& edges = topology().edges();

    forAll(edges, eI)
    {
        const edge& e = edges[eI];

        os << "l " << e.start() + 1 << ' ' << e.end() + 1 << endl;
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// ************************************************************************* //
